/* 
 *
 * fd_select.c:
 *
 * Part of the Form Designer.
 *
 * This file contains all routines and data types to maintain the current
 * selection and manipulate (move, scale) and draw it.
 *
 * Written by Mark Overmars
 *
 * Version 2.1b
 * Date  Nov  3, 1992
 */

#include <gl/gl.h>
#include <gl/device.h>

#include "forms.h"
#include "fd_main.h"

/****
  The Data Structure
****/

#define MAXSEL	2000

FL_OBJECT *selobj[MAXSEL];	/* The selected objects */
int selnumb = 0;		/* Their number */
int backf = FALSE;		/* Whether the selection is the backface */

FL_OBJECT *tmpobj[MAXSEL];	/* A temporary list of objects */
int tmpnumb = 0;		/* Their number */

static int find_selobject(FL_OBJECT *obj)
/* Returns the index of a particular object. -1 if it is not selected */
{
  int i;
  for (i=0; i<selnumb; i++) if (selobj[i] == obj) return i;
  return -1;
}

static void cleanup_selection()
/* Cleans up the selection, ordering the objects and creating
   groups if all elements are in there */
{
  FL_OBJECT *obj, *begobj;
  int i,tt,sel;
  if (cur_form == NULL)
  {
    selnumb = 0;
    backf = FALSE;
    return;
  }
  /* Figure out whether whole groups are selected */
  obj = cur_form->first;
  while (obj != NULL)
  {
    if (obj->objclass == FL_BEGIN_GROUP)
    {
      sel = 1;
      begobj = obj;
    }
    else if (obj->objclass == FL_END_GROUP)
    {
      if (sel)
      {
        selobj[selnumb++] = begobj;
        selobj[selnumb++] = obj;
      }
      else
      {
        if ((tt = find_selobject(begobj)) != -1) selobj[tt] = NULL;
        if ((tt = find_selobject(obj)) != -1) selobj[tt] = NULL;
      }
    }
    else
    {
      if (find_selobject(obj) == -1) sel = 0;
    }
    obj = obj->next;
  }
  /* make a new, ordered list of selected items, removing duplicates
     and NULL objects */
  obj = cur_form->first;
  tmpnumb = 0;
  while (obj != NULL)
  {
    if ((tt = find_selobject(obj)) != -1) tmpobj[tmpnumb++] = obj;
    obj = obj->next;
  }
  for (i=0; i<tmpnumb; i++) selobj[i] = tmpobj[i];
  selnumb = tmpnumb;
  fillin_groups();
}

int is_selected(FL_OBJECT *obj)
/* Returns whether object is selected */
  { return (find_selobject(obj) != -1); }

void addto_selection(FL_OBJECT *obj)
/* Adds an object to the current selection */
{
  if (backf) return;		/* Don't add objects with backface */
  if (selnumb == MAXSEL) return;
  selobj[selnumb++] = obj;
  cleanup_selection();
}

void addgroupto_selection(FL_OBJECT *obj)
/* Adds a group to the current selection */
{
  FL_OBJECT *ob;
  if (backf) return;		/* Don't add objects with backface */
  if (obj->objclass != FL_BEGIN_GROUP) return;
  ob = obj;
  while (ob!=NULL && ob->objclass != FL_END_GROUP)
  {
    if (selnumb == MAXSEL) return;
    selobj[selnumb++] = ob;
    ob = ob->next;
  }
  if (ob != NULL) selobj[selnumb++] = ob;
  cleanup_selection();
}

void deletefrom_selection(FL_OBJECT *obj)
/* Removes an object from the selection (if present) */
{
  int ind = find_selobject(obj);
  if (ind != -1) selobj[ind] = NULL;
  cleanup_selection();
}

void deletegroupfrom_selection(FL_OBJECT *obj)
/* Removes a group to the current selection */
{
  FL_OBJECT *ob;
  int ind;
  if (backf) return;		/* Don't remove objects with backface */
  if (obj->objclass != FL_BEGIN_GROUP) return;
  ob = obj;
  while (ob!=NULL && ob->objclass != FL_END_GROUP)
  {
    ind = find_selobject(ob);
    if (ind != -1) selobj[ind] = NULL;
    ob = ob->next;
  }
  cleanup_selection();
}

void clear_selection()
/* Clears the complete selection */
{
  selnumb = 0;
  backf = FALSE;
  cleanup_selection();
}

/****
  Help procedures
****/

static void compute_selbox(float *x,float *y,float *w,float *h)
/* Compute the bounding box of the selection */
{
  int i;
  float x1=100000.0, y1=100000.0, x2=0.0, y2=0.0;
  for (i=0; i<selnumb; i++)
  {
    if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		 selobj[i]->objclass != FL_END_GROUP)
    {
      if (selobj[i]->x < x1) x1 = selobj[i]->x;
      if (selobj[i]->y < y1) y1 = selobj[i]->y;
      if (selobj[i]->x + selobj[i]->w > x2) x2 = selobj[i]->x + selobj[i]->w;
      if (selobj[i]->y + selobj[i]->h > y2) y2 = selobj[i]->y + selobj[i]->h;
    }
  }
  *x = x1; *y = y1; *w = x2-x1; *h = y2-y1;
}

static void find_mousepos(float *mx, float *my)
/* Find position of the mouse mouse */
{
  if (cur_form == NULL) return;
  winset(main_window);
  get_mouse_pos(mx,my);
}

static FL_OBJECT *find_mouseobj()
/* Returns the object under the mouse. */
{
  float xx,yy;
  if (cur_form == NULL) return NULL;
  find_mousepos(&xx,&yy);
  return fl_find_last(cur_form,FL_FIND_MOUSE,xx,yy);
}

/****
  Drawing routines
****/

#define HS	9.0

int hidden = FALSE;

void draw_selbox()
/* Draw the selection box */
{
  float x,y,w,h;
  int i;
  if (selnumb == 0) return;
  /* draw object boxes */
  fl_color(1);
  for (i=0; i<selnumb; i++)
  {
    if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		 selobj[i]->objclass != FL_END_GROUP)
      rect(selobj[i]->x,selobj[i]->y,
		selobj[i]->x+selobj[i]->w-1.0,selobj[i]->y+selobj[i]->h-1.0);
  }
  if (hidden) return;
  /* draw the total box */
  compute_selbox(&x,&y,&w,&h);
  if (!backf) {x -= 1.0; y -= 1.0, w += 2.0; h += 2.0;}
  fl_color(1);
  rect(x,y,x+w-1.0,y+h-1.0);
  rectf(x,y,x+HS-1.0,y+HS-1.0);
  rectf(x+w-HS,y,x+w-1.0,y+HS-1.0);
  rectf(x+w-HS,y+h-HS,x+w-1.0,y+h-1.0);
  rectf(x,y+h-HS,x+HS-1.0,y+h-1.0);
}

/****
  Interaction handling
****/

void handle_move()
/* Handles the moving of the selection */
{
  float x,y,w,h,mx,my;
  float ox,oy,ow,oh;
  long xx,yy;
  int i,s;
  if (cur_form == NULL) return;
  winset(main_window);
  s = getbutton(LEFTSHIFTKEY) || getbutton(RIGHTSHIFTKEY); 
  find_mousepos(&mx,&my);
  compute_selbox(&x,&y,&w,&h);
  if (mx<x || mx>x+w || my<y || my >y+h) return;  /* Not in box */
  hidden = TRUE;
  redraw_the_form(0);
  if (backf)
  {
    if (mx >= x+w-HS && my >= y+h-HS)
    {
      getsize(&xx,&yy);
      set_bounding_box(0.0,0.0,(float)xx,(float)yy);
      scale_box(&x,&y,&w,&h);
      cur_form->w = w; cur_form->h = h;
      selobj[0]->w = w; selobj[0]->h = h;
      set_bounding_box(0.0,0.0,cur_form->w,cur_form->h);
    }
  }
  else if (s)
  {
    copy_selection();
    paste_selection();
  }
  else
  {
    ox = x; oy = y; ow = w; oh = h;
    /* show the rubberband box */
    if (mx <= x+HS && my <= y+HS)
      { x += w; y += h; w = -w; h = -h; scale_box(&x,&y,&w,&h);}
    else if (mx <= x+HS && my >= y+h-HS)
      { x += w; w = -w; scale_box(&x,&y,&w,&h);}
    else if (mx >= x+w-HS && my <= y+HS)
      { y += h; h = -h; scale_box(&x,&y,&w,&h);}
    else if (mx >= x+w-HS && my >= y+h-HS)
      { scale_box(&x,&y,&w,&h);}
    else
      move_box(&x,&y,&w,&h,TRUE);
    /* recompute object sizes */
    for (i=0; i<selnumb; i++)
    {
      if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		   selobj[i]->objclass != FL_END_GROUP)
      {
        selobj[i]->x = x + (selobj[i]->x-ox)*w/ow;
        selobj[i]->y = y + (selobj[i]->y-oy)*h/oh;
        if (ow != 0.0) selobj[i]->w *= w/ow;
        if (oh != 0.0) selobj[i]->h *= h/oh;
      }
    }
  }
  hidden = FALSE;
  redraw_the_form(backf);
  changed = TRUE;
}

void handle_select()
/* handles the selection of objects */
{
  int s;
  FL_OBJECT *obj, *mouseobj;
  float x,y,w,h,mx,my;
  float stepsize;
  s = getbutton(LEFTSHIFTKEY) || getbutton(RIGHTSHIFTKEY); 
  if (cur_form == NULL) return;
  mouseobj = find_mouseobj();
  find_mousepos(&mx,&my);
  if (s) /* Shift Push */
  {
    if (mouseobj == NULL || mouseobj == cur_form->first->next) return;
    if (find_selobject(mouseobj) == -1)
      addto_selection(mouseobj);
    else
      deletefrom_selection(mouseobj);
  }
  else
  {
    clear_selection();
    x = mx; y = my; w = 0.0; h = 0.0;
    stepsize = get_step_size();
    set_step_size(0.0);
    scale_box(&x,&y,&w,&h);
    set_step_size(stepsize);
    obj = cur_form->first->next->next;
    while (obj != NULL)
    {
      if (obj->objclass != FL_BEGIN_GROUP && obj->objclass != FL_END_GROUP &&
    	   	obj->x >= x && obj->y >= y && obj->x+obj->w <= x+w &&
       		obj->y+obj->h <= y+h)
        { addto_selection(obj); }
      obj = obj->next;
    }
    if (selnumb == 0)
    {
      if (mouseobj == NULL) return;
      else if ( mouseobj == cur_form->first->next) /* backf object */
        { addto_selection(mouseobj); backf = TRUE;}
      else
        addto_selection(mouseobj);
    }
  }
}

void select_all()
/* Selects all objects in the form. */
{
  FL_OBJECT *obj;
  if (cur_form == NULL) return;
  clear_selection();
  obj = cur_form->first->next->next;
  while (obj != NULL) { selobj[selnumb++] = obj; obj = obj->next; }
  cleanup_selection();
}

/****
  Operations on the selection
****/

void change_selection()
{
  FL_OBJECT *firstobj = NULL;
  int objclass,i;
  if (cur_form == NULL) return;
  if (selnumb == 1)
    change_object(selobj[0],TRUE);
  else
  {
    for (i=0; i<selnumb;i++)
    {
      if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		 selobj[i]->objclass != FL_END_GROUP)
      {
        if (firstobj == NULL)
	  { firstobj = selobj[i]; objclass = selobj[i]->objclass;}
        else if (objclass != selobj[i]->objclass)
	{
	  fl_show_message("Selected objects have different classes","","");
	  return;
	}
      }
    }
    if (firstobj == NULL) return;
    if (!change_object(firstobj,FALSE)) return;
    for (i=0; i<selnumb;i++)
    {
      if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		 selobj[i]->objclass != FL_END_GROUP)
      {
        change_type(selobj[i],firstobj->type);
        set_attribs(selobj[i], firstobj->boxtype,
	 firstobj->col1, firstobj->col2, firstobj->lcol, firstobj->align,
	 firstobj->lsize, firstobj->lstyle, selobj[i]->label);
      }
    }
  }
  changed = 1;
}

void align_selection(int dir)
/* Aligns the objects in the selection */
{
  float x,y,w,h,gap;
  int used[10000], current;
  int i, j;
  if (backf) return; /* Cannot align the backface */
  if (cur_form == NULL) return;
  if (selnumb <= 1) return; /* Nothing to align */
  compute_selbox(&x,&y,&w,&h);

  if (dir == 4)  /* Horizontal equal distance */
  {
    gap = 0.0;
    for (i=0; i<selnumb; i++) gap += selobj[i]->w;
    gap = (w-gap) / (selnumb-1);
    for (i=0; i<selnumb; i++) used[i] = 0;
    for (j=0; j<selnumb; j++) 
    {
      current = -1;
      for (i=0; i<selnumb; i++) if (!used[i])
        if (current == -1 || selobj[i]->x < selobj[current]->x) current = i;
      used[current] = 1;
      selobj[current]->x = x; x += selobj[current]->w + gap;
    }
  }
  else if (dir == 8)  /* Vertical equal distance */
  {
    gap = 0.0;
    for (i=0; i<selnumb; i++) gap += selobj[i]->h;
    gap = (h-gap) / (selnumb-1);
    for (i=0; i<selnumb; i++) used[i] = 0;
    for (j=0; j<selnumb; j++) 
    {
      current = -1;
      for (i=0; i<selnumb; i++) if (!used[i])
        if (current == -1 || selobj[i]->y < selobj[current]->y) current = i;
      used[current] = 1;
      selobj[current]->y = y; y += selobj[current]->h + gap;
    }
  }
  else for (i=0; i<selnumb; i++)
  {
    switch (dir) {
      case 1:	/* Left */
	selobj[i]->x = x;
	break;
      case 2:	/* Center */
	selobj[i]->x = x+ w/2.0 - selobj[i]->w/2.0;
	break;
      case 3:	/* Right */
	selobj[i]->x = x+w - selobj[i]->w;
	break;
      case 5:	/* Bottom */
	selobj[i]->y = y;
	break;
      case 6:	/* Center */
	selobj[i]->y = y+ h/2.0 - selobj[i]->h/2.0;
	break;
      case 7:	/* Top */
	selobj[i]->y = y+h - selobj[i]->h;
	break;
    }
  }
  redraw_the_form(0);
  changed = 1;
}


void show_selection()
/* Shows all objects in the selection. */
{
  int i;
  if (backf) return; /* Cannot show the backface */
  if (cur_form == NULL) return;
  for (i=0; i<selnumb; i++) fl_show_object(selobj[i]);
}

void hide_selection()
/* Hides all objects in the selection. */
{
  int i;
  if (backf) return; /* Cannot hide the backface */
  if (cur_form == NULL) return;
  for (i=0; i<selnumb; i++) fl_hide_object(selobj[i]);
}

void raise_selection()
/* Raises the selected objects */
{
  int i;
  if (backf) return; /* Cannot raise the backface */
  if (cur_form == NULL) return;
  for (i=0; i<selnumb; i++)
  {
    fl_delete_object(selobj[i]);
    fl_add_object(cur_form,selobj[i]);
  }
  changed = 1;
}

void lower_selection()
/* Lowers the selection */
{
  int i;
  if (backf) return;	/* Cannot lower the backface. */
  if (cur_form == NULL) return;
  for (i=selnumb-1; i>=0; i--)
  {
    if (selobj[i] != cur_form->first->next->next)
    {
      fl_delete_object(selobj[i]);
      fl_insert_object(selobj[i],cur_form->first->next->next);
    }
  }
  changed = 1;
}

static FL_OBJECT *bufobj[10000];	/* Buffered objects */
static int bufnumb = 0;			/* Their number */

void cut_selection()
/* Removes all elements in the selection. */
{
  int i;
  if (backf) return;	/* Cannot cut the backface. */
  if (cur_form == NULL) return;
  if (selnumb == 0) return;
  /* Clear the buffer */
/* CAUSES PROBLEMS WITH NAMING
  for (i=0; i<bufnumb; i++) { fl_free_object(bufobj[i]);}
*/
  /* Make new deletion and save it */
  for (i=0; i<selnumb; i++)
    { fl_delete_object(selobj[i]); bufobj[i] = selobj[i]; }
  bufnumb = selnumb;
  clear_selection();
  changed = 1;
}

void paste_selection()
/* Pastes elements form buffer into form */
{
  FL_OBJECT *obj;
  float x,y,w,h,ox,oy;
  int i;
  if (cur_form == NULL) return;
  if (bufnumb == 0) return;
  /* Copy selection from buffer */
  clear_selection();
  redraw_the_form(0);
  for (i=0; i<bufnumb; i++)
  {
    obj = copy_object(bufobj[i]);
    fl_add_object(cur_form,obj);
    selobj[selnumb++] = obj;
  }
  /* Move the selection to the correct place */
  compute_selbox(&x,&y,&w,&h);
  ox = x; oy = y;
  move_box(&x,&y,&w,&h,FALSE);
  /* recompute object sizes */
  for (i=0; i<selnumb; i++)
  {
    if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		 selobj[i]->objclass != FL_END_GROUP)
      { selobj[i]->x += x-ox; selobj[i]->y += y-oy; }
  }
  cleanup_selection();
  redraw_the_form(0);
  changed = 1;
}

void copy_selection()
/* Copies all elements in the selection  to the buffer*/
{
  FL_OBJECT *obj;
  int i;
  if (backf) return;	/* Cannot copy the backface */
  if (cur_form == NULL) return;
  if (selnumb == 0) return;
  /* Clear the buffer */
/* CAUSES PROBLEMS WITH NAMING
  for (i=0; i<bufnumb; i++) { fl_free_object(bufobj[i]);}
*/
  bufnumb = 0;
  /* Copy the objects */
  for (i=0; i<selnumb; i++)
  {
    obj = copy_object(selobj[i]);
    bufobj[i] = obj;
  }
  bufnumb = selnumb;
}

void group_selection()
/* Groups all elements in the selection into a group */
{
  int i;
  FL_OBJECT *obj;
  if (backf) return;	/* Cannot group the backface */
  if (cur_form == NULL) return;
  if (selnumb == 0) return;
  obj = add_an_object(FL_BEGIN_GROUP,-1,0.0,0.0,0.0,0.0);
  for (i=0; i<selnumb; i++)
  {
    {
      fl_delete_object(selobj[i]);
      if (selobj[i]->objclass != FL_BEGIN_GROUP &&
		selobj[i]->objclass != FL_END_GROUP)
        fl_add_object(cur_form,selobj[i]);
    }
  }
  add_an_object(FL_END_GROUP,-1,0.0,0.0,0.0,0.0);
  clear_selection();
  set_object_name(obj,fl_show_input("Give group name",""),"","");
  addgroupto_selection(obj);
  changed = 1;
}

void flatten_selection()
/* Removes begin and endgroups from the selection */
{
  int i;
  if (backf) return;	/* Cannot flatten the backface */
  if (cur_form == NULL) return;
  for (i=0; i<selnumb; i++)
  {
    if (selobj[i]->objclass == FL_BEGIN_GROUP ||
		 selobj[i]->objclass == FL_END_GROUP)
    {
      fl_delete_object(selobj[i]);
/* CAUSES PROBLEMS WITH NAMING
      fl_free_object(selobj[i]);
*/
      selobj[i] = NULL;
    }
  }
  cleanup_selection();
  changed = 1;
}
